"""
    Animation Optimizer
"""

import os
import sys
import argparse
import math
from io import StringIO
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from glm import length

import torchvision
from torchvision import datasets, transforms
from torch.autograd import Variable
from torchsummary import summary

from pydeepan import log
from pydeepan import dataset
from pydeepan.chara.HPASkeleton import HPASkeleton
from pydeepan.chara.HPAAnimation import HPAAnimation




class AnimationOptimizer:
    """
        A base class for optimizing animation
        Data
            optimizer: pytorch optimizer, it will be initialized in the 'optimize' function
            device: pytorch device
    """
    def __init__(self, device=torch.device("cpu")):
        print( "AnimationOptimizer::__init__" )
        self.optimizer = None
        self.device = device


    def loss(self):
        return 0.0


    def optimize(self, src:HPAAnimation, n:int)->HPAAnimation:
        print("AnimationOptimizer::optimize...")
        return src




#######################################################################################################################################

class AOSkeletonConstraints(AnimationOptimizer):
    """
        Optimize/Re-adjust the limb length contraints.
        This exemple optimizes the skeleton limb lengths.
    """

    def __init__(self, anim:HPAAnimation, device=torch.device("cpu")):
        """
            skeleton (HPASkeleton): init as an HPASkeleton
            limb_length(tensor): init with None, it will be init at the first call of optimize
        """
        print( "AOSkeletonConstraints::__init__: ..." )
        super().__init__(device)
        self.skeleton = HPASkeleton()

        ### compute limb lengths
        anim_np = np.resize( anim, (HPAAnimation.n_frames, HPAAnimation.dim_pose+2))        # resize pour etre un multiple de 3
        anim_tensor = torch.tensor( anim_np ).to(self.device)
        self.limb_lengths = self.skeleton.compute_limb_lengths( anim_tensor  )

        ### just check that limb are well calculated in the compute_limb_lengths function
        # f = 0
        # for i in range(20):
        #     pid = self.skeleton[i+1].parent.index
        #     l1 = length( anim.joint_pos(f,i+1) - anim.joint_pos(f,pid) )
        #     l2 = self.limb_lengths[f,i]
        #     print("l="+str(l1) + " "+str(l2))
        # print( str(self.limb_lengths) )
        print( "AOSkeletonConstraints::__init__: ok" )


    def loss(self, dst_tensor):
        limb_lengths = self.skeleton.compute_limb_lengths( dst_tensor  )
        lo = torch.mean((limb_lengths - self.limb_lengths)**2)
        return lo


    def optimize(self, src:HPAAnimation, n:int, print_loss=False)->HPAAnimation:
        """
         optimize source animation (src), store the result in destination animation (dst)
         src (numpy)
        """
        src_np = np.resize( src, (src.n_frames,75))     # 
        result_tensor = torch.tensor( src_np ).to(self.device).requires_grad_(True)

        if self.optimizer == None:
            self.optimizer = optim.SGD( [result_tensor] , lr=100.0, momentum=0.5)
        #optimizer = optim.Adam([target], lr=0.003)

        for i in range(n):
            # the loss
            total_loss = self.loss( result_tensor )

            # one step: update your target animation
            self.optimizer.zero_grad()
            total_loss.backward()
            self.optimizer.step()
            
            #if  i % show_every == 0:
            if print_loss:
                print('Total loss: ', i, total_loss.item())

        result = result_tensor.to("cpu").clone().detach().numpy()
        result = np.resize( result, (src.n_frames,src.dim_pose))
        result = HPAAnimation.cast( result )
        return result





if __name__ == "__main__":
    print("END")
    #anims = np.load( dataset.dir_anims_processed_npz + 'holden/data_styletransfer.npz')
    anims = np.load( dataset.name2filename("styletransfer") )
    #anims = np.load( dataset.name2filename("emilya") )
    #anims = np.load( dataset.name2filename("cmu") )
    anims = anims['clips']
    anim_hpa = HPAAnimation.cast( anims[0] )
    aoptim = AOSkeletonConstraints( anim_hpa )
    anim_hpa.perturb()
    anim_o = aoptim.optimize( anim_hpa, 100, print_loss=True)
