import typing
from random import random
import numpy as np
from glm import vec3

from pydeepan import dataset


# ==========================================================================================================
# ** Holden's format (in the NPZ files) : 73 values = 22 Positions(XYZ) + VX, VZ, VangY + 4 contacts
#                                                       FatherID
#  0 ** 0, 1, 2 : projection du root sur le sol         -1 (father) 
#  1 ** 3, 4, 5 : root; vertebre basse                  -1
#  2 ** 6, 7, 8 : hanche gauche                         1
#  3 ** 9, 10, 11 : genou gauche                        2
#  4 ** 12, 13, 14 : cheville gauche                    3
#  5 ** 15, 16, 17 : orteils gauche                     4
#  6 ** 18, 19, 20 : hanche droite                      1
#  7 ** 21, 22, 23 : genou droite                       6
#  8 ** 24, 25, 26 : cheville droite                    7
#  9 ** 27, 28, 29 : orteils droite                     8
# 10 ** 30, 31, 32 : vertebre milieu                    1
# 11 ** 33, 34, 35 : vertebre haute                     10
# 12 ** 36, 37, 38 : nuque                              11
# 13 ** 39, 40, 41 : tete                               12
# 14 ** 42, 43, 44 : epaule gauche                      12
# 15 ** 45, 46, 47 : coude gauche                       14
# 16 ** 48, 49, 50 : poignet gauche                     15
# 17 ** 51, 52, 53 : doigts gauche                      16
# 18 ** 54, 55, 56 : epaule droite                      12
# 19 ** 57, 58, 59 : coude droite                       18
# 20 ** 60, 61, 62 : poignet droite                     19
# 21 ** 63, 64, 65 : doigts droite                      20
# xx ** 66 : Vitesse X
# xx ** 67 : Vitesse Z
# xx ** 68 : Vitesse angulaire autour de Y
# xx ** 69, 70 : Contact pied gauche
# xx ** 71, 72 : Contact pied droit
# ==========================================================================================================


class HPAAnimation(np.ndarray):
    """
        wrapper class from ndarray containing HPA animation: 240 frames x 73 values (see below)
    """

    # def n_frames(self)->int:
    #     assert self.shape[0] == 240
    #     return 240

    n_frames = 240
    dim_pose = 73
    n_joints = 21

    # @property
    # def n_joints(self)->int:
    #     return 21

    # @property
    # def dim_pose(self)->int:
    #     assert self.shape[1] == 73
    #     return 73

    @classmethod
    def cast(cls, some_anim: np.ndarray):
        return some_anim.view( HPAAnimation )


    def joint_pos(self, frame:int, jid:int)->vec3:
        return self.__getitem__( (frame, jid+1) )


    def __getitem__(self, tup)->vec3:
        frame_id, joint_id = tup
        assert frame_id <= self.n_frames and frame_id >= 0
        assert joint_id <= self.n_joints and joint_id >= 0

        if joint_id<=21:
            x = vec3( 
                            super().__getitem__( (frame_id, joint_id*3)   ),
                            super().__getitem__( (frame_id, joint_id*3+1) ),
                            super().__getitem__( (frame_id, joint_id*3+2) )
                    )
        elif joint_id>=66 and joint_id<=72:
            x = super().__getitem__( (frame_id, joint_id) ),
        else:
            print("HPAAnimation::__getitem__: error")
            raise
        return x


    def velocity(self, frame_id:int)->vec3:
        return vec3(    super().__getitem__( (frame_id,66) ), 
                        0, 
                        super().__getitem__( (frame_id,67) ) 
                    )


    def perturb(self):
        """
            perturb animations, used to test lenght constraint optimization
        """
        for f in range(self.n_frames):
            for j in range(self.dim_pose):
                v = super().__getitem__( (f,j) )
                v = v + 0.9 * random() 
                super().__setitem__( (f,j), v )




if __name__ == "__main__":

    #stylet = np.load( dataset.name2filename("styletransfer") )
    emilya = np.load( dataset.name2filename("emilya") )
    cmu = np.load( dataset.name2filename("cmu") )
    #mpi = np.load( dataset.name2filename("MPI") )

    #n = stylet['clips'].shape[0]
    #n=5
    nd=[1000,0]
    n = np.sum(nd)
    print("n="+str(n) + " nd="+str(nd))

    emilya_clips = emilya['clips']
    emilya_classes = emilya['classes']
    cmu_clips = cmu['clips']

    w,h = emilya_clips[0].shape
    print("shape "+str(w)+" "+str(h))

    anims = np.empty( [n,w,h], dtype=float )
    label = np.empty( [n,1,], dtype=int )      # keep only style (do not take action)
    ### add emilya
    for i in range(0, nd[0], 1):
        id  = np.random.randint(emilya_clips.shape[0])
        #print("id="+str(id))
        anims[i] = emilya_clips[id]
        label[i] = emilya_classes[id][1]
        print("i="+str(i) + "  id="+str(id))

    ### add cmu
    for i in range(nd[0], nd[0]+nd[1], 1):
        id  = np.random.randint(cmu_clips.shape[0])
        anims[i] = cmu_clips[id]
        label[i] = 7                               # 7 is neutral
        print("i="+str(i) + "  id="+str(id))

    ### shake

    ### save
    npzname = dataset.name2filename("mixing")
    print("filename="+str(npzname))
    print("clips shape="+str(anims.shape))
    print("classes/labels shape="+str(label.shape))
    np.savez_compressed( npzname, clips=anims, classes=label)



    print("anims  shape="+str(anims.shape) + " type="+ str(type(anims)))


    #anims = np.load( dataset.dir_anims_processed_npz + 'holden/data_styletransfer.npz')
    anims = np.load( dataset.name2filename("styletransfer") )
    #anims = np.load( dataset.name2filename("emilya") )
    #anims = np.load( dataset.name2filename("cmu") )
    anims = anims['clips']
    print("anims  shape="+str(anims.shape) + " type="+ str(type(anims)))

    print("============ before cast ")
    anim = anims[0]
    print( "anim: shape="+ str(anim.shape) + " type="+str(type(anim)) )
    print("anim: pose[10,4]="+str(anim[10,4]))


    anim_hpa = anim.view( HPAAnimation )
    anim_hpa = HPAAnimation.cast( anim )

    print("============ after cast ")
    print( type(anim_hpa) )
    print("anim_hpa: frame 10, root="+str(anim_hpa[10,1]))
    print("velocity=" + str(anim_hpa.velocity(0)))

    anim_hpa[10,1].x = anim_hpa[10,1].x+1
    print("anim_hpa: frame 10, root="+str(anim_hpa[10,1]))
    print("anim: pose[10,4]="+str(anim[10,4]))

    anim[10,4] = anim[10,4] + 1
    print("anim_hpa: frame 10, root="+str(anim_hpa[10,1]))
    print("anim: pose[10,4]="+str(anim[10,4]))

    print("dim=" + str(anim_hpa.n_frames) + " " + str(HPAAnimation.n_frames) )

#    print("anim_hpa[0,0]="+str(anim_hpa[0,0]))
#    print("anim[0,0]="+str(anim[0,0]))


    print("END")
