import typing
import os, sys, copy
import numpy as np

from multipledispatch import dispatch

from pydeepan.chara.Skeleton import Joint
from pydeepan.chara.Skeleton import Skeleton
from pydeepan.math.Transform import Transform
from pydeepan.dataset.BVHReader import BVHReader
from pydeepan.chara.HPAAnimation import HPAAnimation



class HPASkeleton(Skeleton):
    """
        A skeleton in the format of HPA (Holden Posotion Animation, see README)
    """
    def __init__(self):
        super().__init__( )

        # Build HPA skeleton
        root = self.add_joint( Joint("root",None)  )             # 0
        lhip = self.add_joint( Joint("left_hip", root) )        # 1
        lkne = self.add_joint( Joint("left_knee", lhip) )       # 2
        lhen = self.add_joint( Joint("left_hankle", lkne) )     # 3
        ltor = self.add_joint( Joint("left_toes", lhen) )       # 4
        rhip = self.add_joint( Joint("right_hip", root) )       # 5
        rkne = self.add_joint( Joint("right_knee", rhip) )      # 6
        rhan = self.add_joint( Joint("right_hankle", rkne) )    # 7
        rtoe = self.add_joint( Joint("right_toes", rhan) )      # 8
        spi1 = self.add_joint( Joint("spine_1", root) )         # 9
        spi2 = self.add_joint( Joint("spine_2", spi1) )         # 10
        neck = self.add_joint( Joint("neck", spi2) )            # 11
        head = self.add_joint( Joint("head", neck) )            # 12
        lsho = self.add_joint( Joint("left_shoulder", neck) )   # 13
        lelb = self.add_joint( Joint("left_elbow", lsho) )      # 14
        lwri = self.add_joint( Joint("left_wrist", lelb) )      # 15
        lfin = self.add_joint( Joint("left_fingers", lwri) )    # 16
        rsho = self.add_joint( Joint("right_shoulder", neck) )  # 17
        relb = self.add_joint( Joint("right_elbow", rsho) )     # 18
        rwri = self.add_joint( Joint("right_wrist", relb) )     # 19
        rfin = self.add_joint( Joint("right_fingers", rwri) )   # 20

        
    @dispatch(np.ndarray, int)
    def set_pose(self, anim:np.ndarray, frame_id:int):
        self.set_pose(  HPAAnimation.cast(anim), frame_id )

    @dispatch(BVHReader, int)
    def set_pose(self, anim:BVHReader, frame_id:int):
        super().set_pose(self, anim, frame_id)

    @dispatch(HPAAnimation, int)
    def set_pose(self, anim:HPAAnimation, frame_id:int):
        """ anim may be HPAAnimation or BVHReader or any other defined in derivative class"""
        #if type(anim) == HPAAnimation:
        for i in range(self.n_joints):      # n_joints should be 21
            j = self.joints[i]
            j.setPos( anim[frame_id, i+1] )         # i+1 since the Joint 0 is the projection of the root node on the ground in HPAAnimation



if __name__ == "__main__":
    hpa_ske = HPASkeleton()
    print(str(hpa_ske))