import os, sys, copy

from multipledispatch import dispatch

from glm import vec3

from pydeepan.dataset.BVHReader import BVHReader
from pydeepan.dataset import HPASkeleton_BVH_Filename
from pydeepan.math.Transform import Transform
from pydeepan.chara.HPAAnimation import HPAAnimation

import torch



class Joint:
    """
        Joint node of a Skeleton (see class below)
    """
    def __init__(self, name: str, parent = None):
        self.name = name
        self.index = -1                         # index in the joint array of the skeleton
        self.parent = parent
        self.children = []
        self.transform = Transform()            # identity


    @property
    def has_parent(self):
        return self.parent != None


    @property
    def parent_index(self)->int:
        if not self.is_root:
            return self.parent.index
        else:
            return -1

    @property
    def is_root(self)->bool:
        return self.parent == None


    @property
    def is_EndSite(self)->bool:
        return len(self.children)==0


    def setPos(self, pos:vec3):
        self.transform.setTranslation( pos )


    def __str__(self)->str:
        t = "joint("
        t += str(self.index) + "/"
        t += "\"" + self.name+"\"; "
        t += "parent="
        if self.parent != None:
            t += str(self.parent.index) + "/\"" + self.parent.name + "\""
        else:
            t += "None"
        t += "; #children=" + str( len(self.children))
        t += ")"
        return t







class Skeleton:
    """
        Skeleton is a Hierarchy of joints

          - joints([]) of Joint
          - _joints_indices(dict): indice from joint name
    """
    def __init__(self, bvh_filename: str = None):
        self.joints = []                            # joints are store like in a tree path in width
        self._joints_indices = {}                   # dict from joint name

        if bvh_filename != None:
            bvh = BVHReader( bvh_filename )
            bvh.read( includeEndSite=False )
            print(str(bvh))
            for j in range(len(bvh._nodes)):
                node = bvh._nodes[j]
                self.add_joint( Joint( node.name, node.parent_id) )



    def add_joint(self, joint: Joint, parent=None)->int:
        """
            add a joint. parent may be int, Joint, None
        """
        if parent is None:
            parent = joint.parent
        if parent==None:
            assert len(self.joints)==0
        else:
            if type(parent) == int: # parent is an index
                if parent==-1:
                    joint.parent = None
                else:
                    joint.parent = self.joints[parent]
            elif type(parent) == str: # parent is a name
                joint.parent = self.get_joint_by_name(parent)
            elif type(parent) == Joint:
                joint.parent = parent
            if joint.parent != None:
                joint.parent.children.append(joint)
        
        joint.index = len(self.joints)
        self._joints_indices[joint.name] = joint.index
        self.joints.append( copy.deepcopy(joint) )
        return joint.index


    @property
    def root_joint(self)->Joint:
        return self.joints[0]


    @property
    def n_joints(self)->int:
        # if self.root_velocity: 
        #     return len(self.joints) - 1
        # else: 
            return len(self.joints)


    @property
    def parents(self):
        """Returns an list of each joint's parent"""
        return [j.parent for j in self.joints]


    @property 
    def parents_indices(self):
        """Returns an list of parent's indice"""
        res = []
        for j in self.joints:
            if j.parent:
                res.append(j.parent.index)
            else:
                res.append(None)
        return res


    @property
    def dimensions(self)->int:
        return len(self.joints) * 3
    

    @property
    def limbs_count(self)->int:
        # count = 0
        # for j in self.joints:
        #     if j.parent is not None: count+=1
        # return count
        return len(self.joints)-1
    

    def get_joint_by_index(self, idx: int)->Joint:
        return self.joints[idx]


    def get_joint_by_name(self, name: str)->Joint:
        """Return the joint with the given name"""
        return self.joints[self._joints_indices[name]]


    def __getitem__(self, idx: int)->Joint:
        return self.joints[idx]
    

    def compute_limb_lengths(self, x:torch.tensor)->torch.tensor:
        """ 
            x (tensor): tensor representing an HPAAnimation
            return tensor of limb length 
        """
        b, f = x.size()
        assert b == HPAAnimation.n_frames
        assert f == HPAAnimation.dim_pose+2  # tensor has to be expanded to be a multiple of 3
        x = x.view(b, -1, 3)
        lengths = torch.empty((b, self.limbs_count), device=x.device)

        idx=0
        for i, j in enumerate(self.joints):
            if j.parent is not None:
                lengths[:, idx] = (x[:, j.index+1] - x[:, j.parent.index+1]).norm(dim=-1)         # +1 since the first line of x is the root projected on ground
                idx += 1 # only update index if the current joint has an index (do not take into account the root)
                # if idx == 0:
                #     print( "i_tab="+str(j.index+1) + " parent_tab="+str(j.parent.index+1) 
                #             + "vec3="+str(x[0,j.index+1]) 
                #             + "vec3_parents="+str(x[0,j.parent.index+1]) 
                #         )
        return lengths


    def __str__(self)->str:
        t = "Skeleton: "+str(len(self.joints))+"\n"
        #for j in self.joints:
        for x in range(len(self.joints)):
            t += "   " + str(x) + ": "+ str(self.joints[x]) + "\n"
        return t


    @dispatch(object, int)
    def set_pose(self, anim, frame_id:int):
        print("Skeleton::set_pose: anim is generic")


    @dispatch(BVHReader,int)
    def set_pose(self, anim:BVHReader, frame_id:int):
        """ anim may be BVHReader or any other define in derivative class"""
        #if type(anim) == BVHReader:
        print("TODO")





if __name__ == "__main__":
    print("=========================================== Skeleton")
    ske = Skeleton()
    root = ske.add_joint( Joint("root", None), )
    shoulder = ske.add_joint( Joint("shoulder", root), )
    elbow = ske.add_joint( Joint("elbow", shoulder), )
    wrist = ske.add_joint( Joint("wrist", elbow), )
    wrist2 = ske.add_joint( Joint("wrist2", elbow), )
    print( str(ske) )

    print("=========================================== Skeleton from BVH")
    print(HPASkeleton_BVH_Filename)
    ske = Skeleton( HPASkeleton_BVH_Filename )
    print( str(ske) )

    print("=========================================== set_pose")

    bvh = BVHReader( HPASkeleton_BVH_Filename )
    bvh.read( includeEndSite=False )
    print(str(bvh))

    ske.set_pose( "str", 0)
    ske.set_pose( bvh, 0)


