"""
HPADataset: manage a dataset of HPA (Holden Position Animation). 
The main here loads all known datasets and print some informations.
See HPADatasetDemo.py for a 3D Viewer demo.
"""


import pydeepan
from pydeepan import log, tools
from pydeepan.torch_import import *
from pydeepan import dataset as AnimDataset



class HPANormalize:
    '''
        A class that containt mean + std deviation to preprocesse HPAAnimation
        Arguments
            - Xmean(np array of float): mean
            - Xstd(np array of float): standard deviation
            - filename(str): filename 
    '''

    def __init__(self, dataset_name:str, X:np.array=None):
        '''
        provide an np.ndarray(optional) + a filename (or a dataset name)
        '''
        assert dataset_name!=None
        if X is not None:
            self.compute( dataset_name, X)
            self.save( self.filename )
        else:
            self.load( dataset_name )


    def _set_filename(self, filename):
        if not "_meanstd" in filename:
            filename = AnimDataset.name2filename(filename)
            self.filename = tools.add_to_filename(filename, "_meanstd")
        else:
            self.filename = filename


    def compute(self,  filename:str, X:np.array):
        self._set_filename( filename )
        X = np.swapaxes( X, 1, 2) #.astype(np.float64)
        feet = np.array([12,13,14,15,16,17,24,25,26,27,28,29])
        Xmean = X.mean(axis=2).mean(axis=0)[np.newaxis,:,np.newaxis]
        Xmean[:,-7:-4] = 0.0
        Xmean[:,-4:]   = 0.5
        Xstd = np.array([[[X.std()]]]).repeat(X.shape[1], axis=1)
        Xstd[:,feet]  = 0.9 * Xstd[:,feet]
        Xstd[:,-7:-5] = 0.9 * X[:,-7:-5].std()
        Xstd[:,-5:-4] = 0.9 * X[:,-5:-4].std()
        Xstd[:,-4:]   = 0.5
        self.Xmean = Xmean
        self.Xstd = Xstd


    def preprocess(self, anims):
        """ take a numpy=batch of animations, preprocess it and return a numpy """
        X = np.swapaxes( anims, 1, 2) #.astype(np.float64)
        X = (X - self.Xmean) / self.Xstd
        return X


    def postprocess(self, anims):    
        """ take a numpy=batch of animations, postprocess them and return a numpy """
        #Xorgi = (Xorgi * preprocess['Xstd']) + preprocess['Xmean']
        X = (anims * self.Xstd) + self.Xmean
        X = np.swapaxes( X, 1, 2)
        return X


    def save(self, filename:str=None):
        if filename != None:
            self._set_filename( filename )
        tools.create_dir(self.filename)
        np.savez_compressed( self.filename, Xmean=self.Xmean, Xstd=self.Xstd)
        print("HPANormalize: save preprocessing data (mean, std)..."+self.filename)


    def load(self, meanstd_filename:str):
        # try:
        #     if self.preprocessed:
        #            print("WARNING New preprocessing data (mean,std) are loaded but the dataset was already preprocessed with different mean/std !!!")
        #            self.clips = self.postprocess( self.clips )
        # except: pass
        self._set_filename( meanstd_filename )
        mean_std = np.load( self.filename )
        # try:
        #     diff = mean_std['Xmean'].all != self.Xmean.all or mean_std['Xstd'].all != self.Xstd.all
        # except: pass
        self.Xmean = mean_std['Xmean']
        self.Xstd = mean_std['Xstd']
        print("HPANormalize: load preprocessing data (mean,std)..."+self.filename)
        # try:
        #     if self.preprocessed:
        #            print("WARNING Repreprocess all data with new mean/std ! new values are differents?="+str(diff))
        #            self.clips = self.preprocess( self.clips )
        # except: pass


    def mean_std(self)->np.array:
        return np.array( Xmean=self.Xmean, Xstd=self.Xstd)


    def __str__(self):
        res = "HPANormalize: name="+str(self.filename)+"  Xmean.shape="+str(self.Xmean.shape)+ " Xstd.shape="+str(self.Xstd.shape) + "\n"
        return res





class HPADataset(Dataset):      
    """
    # Class to manage a dataset of HPA animation for PyTorch computation (tensor)
    # HPA for Holden Positional Animation
    """

    def __init__(self, dataset_filename, transform=None, subpart=-1.0):      # HPA dataset, if meanstd_filename exist: computed extra data(mean, std) else load specific mean/std data related to an already trained autoencoder
        """
            file(str): npz filename (full file name or shot file name like 'cmu' or 'emilya')
            transform(array): array of preprocessing to apply to each sample (See HPATransformClipWithoutLabel, etc.)
            subpart(float): to cut dataset into two parts, training dataset and validation dataset for instance.
        """
        super(HPADataset, self).__init__()
        self.transform=transform
        file = AnimDataset.name2filename(dataset_filename)
        npz = np.load( file, allow_pickle=True )
        self.filename = file
        print("HPADataset: file=%s    data names of array included=%s" % (file, str(list(npz.keys()))))
        X = npz['clips']
        if abs(subpart)<1.0:
            print("HPADataset: clips shape before subpart="+str(X.shape))
            if subpart<0:
                last = int(X.shape[0]*-subpart)
                X = X[:last]
            else:
                start = int(X.shape[0]*subpart)
                X = X[start:]
            print("HPADataset: clips shape after subpart="+str(X.shape))

        self.hpanorm = HPANormalize( file, X )
        X = self.hpanorm.preprocess(X)
        self.clips = X
        self.preprocessed = True                # use the mean, std normalization or not
        #print("HPADataset: clips.shape=" + str(self.clips.shape) + " Xmean.shape"+str(self.Xmean.shape)+ " Xstd.shape"+str(self.Xstd.shape))

        self.nbr_classes = 0
        if 'classes' in npz.keys():
            self.classes = npz['classes']
            if self.classes.ndim>1:
                self.nbr_classes = self.classes.shape[1]
            else:
                self.nbr_classes = 1
            print("HPADataset: classes is present. Number of classes(if 1=>style, if 2=>(action,style))="+str(self.nbr_classes))
            if abs(subpart)<1.0:
                print("HPADataset: classes.shape before subpart=" + str(self.classes.shape))
                if subpart<0:
                    last = int(self.classes.shape[0]*-subpart)
                    self.classes = self.classes[:last]
                else:
                    start = int(self.classes.shape[0]*subpart)
                    self.classes = self.classes[start:]
                print("HPADataset: classes.shape after subpart=" + str(self.classes.shape))
        else:
            self.classes = None


    def __str__(self):
        res = "HPADataset: filename=" + self.filename + "\n"
        res += "===>clips.shape=" + str(self.clips.shape)
        res += "===>"+ str(self.hpanorm)
        res += "===>preprocessed=" + str(self.preprocessed) + "\n"
        res += "===>nb of classes=" + str(self.nbr_classes) + "(1=clip+style; 2=clip,action,style; 0=clip)" + "\n"
        res += "===>dim labels style: "+str(self.dim_labels_style())+ "\n"
        return res


    def __len__(self):
        return len(self.clips)


    def get_brut_clip(self, idx):
        """ return clip  postprocessed, so ready to be displayed """
        c = self.clips[idx]
        c = c.reshape( [1, c.shape[0], c.shape[1] ] )
        X = self.hpanorm.postprocess( self.clips[idx] )             # since posprocess take a batch
        X = X.reshape( [X.shape[1], X.shape[2] ] )
        return X


    def __getitem__(self, idx):     
        """ return the idx clip, preprocessed and transformed to torch tensor (ready to go through a network) """
        clip = self.clips[idx]
        if not self.preprocessed:
            clip = self.hpanorm.preprocess(clip)
        clip = self.to_torch_tensor_float( clip )
        if self.nbr_classes==1:
            classes = self.classes[idx]
            classes =  torch.from_numpy(classes).long()
            sample = { 'clip': clip, 'style': classes[0], 'id':idx }
        elif self.nbr_classes==2:
            classes = self.classes[idx]
            classes =  torch.from_numpy(classes).long()
            sample = { 'clip': clip, 'action': classes[0], 'style': classes[1], 'id': idx }
        else:
            sample = { 'clip': clip, 'id': idx }
        if self.transform!=None:
            sample = self.transform(sample)
        return sample


    def get_shapes(self):
        """ return the 3 shapes: input_sample, preprocessed_data, number of classes (1=style; 2=style,action) """
        sample = self.__getitem__(0)
        clipbrut = self.get_brut_clip(0)
        if self.nbr_classes==0:
            return clipbrut.shape, sample.shape, self.nbr_classes
        else:
            return clipbrut.shape, sample[0].shape, self.nbr_classes


    def dim_labels(self, axis):
        #return (np.amax( self.classes[:,axis] ) - np.amin( self.classes[:,axis] ))+1
        # if self.classes.ndim==1:
        #     return (np.amax( self.classes[axis])  - 0 )+1
        # else:
        if self.classes is None:
            return 0
        else:
            return (np.amax( self.classes[:,axis])  - 0 )+1


    def dim_labels_style(self):
        if self.classes is None:
            return 0
        else:
            return self.dim_labels( self.classes.shape[1]-1 )       # last clolumn is the style

    @staticmethod
    def to_torch_tensor_float(sample):
        return torch.from_numpy(sample).float()

    @staticmethod
    def to_torch_tensor_long(sample):
        return torch.from_numpy(sample).long()

    # def mean_std(self):
    #     return np.array( Xmean=self.Xmean, Xstd=self.Xstd)

    # def preprocess(self, anim):
    #     """ take a numpy=batch of animations, preprocess it and return a numpy """
    #     X = np.swapaxes( anim, 1, 2) #.astype(np.float64)
    #     X = (X - self.Xmean) / self.Xstd
    #     return X


    # def postprocess(self, anims):    
    #     """ take a numpy=batch of animations, postprocess them and return a numpy """
    #     #Xorgi = (Xorgi * preprocess['Xstd']) + preprocess['Xmean']
    #     X = (anims * self.Xstd) + self.Xmean
    #     X = np.swapaxes( X, 1, 2)
    #     return X


    # def meanstd_filename_from_hpadataset(self, hpa_filename):
    #     #hpa_name, ae_ext = os.path.splitext(hpa_filename)
    #     #filename = hpa_filename.replace( hpa_name, hpa_name + "_meanstd" )
    #     return tools.add_to_filename(hpa_filename, "_meanstd")


    # def save_meanstd(self, meanstd_filename):
    #     tools.create_dir(meanstd_filename)
    #     np.savez_compressed( meanstd_filename, Xmean=self.Xmean, Xstd=self.Xstd)
    #     print("Save preprocessing data (mean, std)..."+meanstd_filename)


    # def load_meanstd(self, meanstd_filename):
    #     try:
    #         if self.preprocessed:
    #                print("WARNING New preprocessing data (mean,std) are loaded but the dataset was already preprocessed with different mean/std !!!")
    #                self.clips = self.postprocess( self.clips )
    #     except: pass
    #     mean_std = np.load( meanstd_filename )
    #     try:
    #         diff = mean_std['Xmean'].all != self.Xmean.all or mean_std['Xstd'].all != self.Xstd.all
    #     except: pass
    #     self.Xmean = mean_std['Xmean']
    #     self.Xstd = mean_std['Xstd']
    #     print("load preprocessing data (mean,std)..."+meanstd_filename)
    #     try:
    #         if self.preprocessed:
    #                print("WARNING Repreprocess all data with new mean/std ! new values are differents?="+str(diff))
    #                self.clips = self.preprocess( self.clips )
    #     except: pass




class HPATransformClipWithoutLabel(object):
    def __call__(self, x):
        #print( "HPATransformOnlyClip len=" + str(len(x)) )
        if len(x)==0:
            return x
        else:
            #print("shape"+str(x['clip'].shape))
            return x['clip']



class HPATransformClipStyle(object):
    def __call__(self, x):
        #print( "HPATransformOnlyClip len=" + str(len(x)) )
        #print( x['clip'])
        if len(x)==0:
            raise ValueError("Style is not present in the data")
        elif len(x)==1:
            return x
        else:
            clip = x['clip']
            style = x['style']
            return clip, style



class HPATransformClipActionStyle(object):
    def __call__(self, x):
        #print( "HPATransformOnlyClip len=" + str(len(x)) )
        #print( x['clip'])
        if len(x)==0:
            raise ValueError("Style is not present in the data")
        elif len(x)==1:
            return x
        else:
            clip = x['clip']
            action = x['action']
            style = x['style']
            return clip, action, style







if __name__ == "__main__":

    log_file = pydeepan.dir_pydeepan + "data/hpadataset.log"
    with log.create_logger(log_file, True):
        # log = log.create_logger(log_file)

        print("##################################")
        hpanorm = HPANormalize("emilya")
        print("hpanorm emilya="+str(hpanorm))

        hpanorm = HPANormalize("mixing")
        print("hpanorm mixing="+str(hpanorm))

        # iterate on dataset_known
        for dataset_name in AnimDataset.dataset_known:
            # Load dataset
            anims_npz_filename = AnimDataset.name2filename(dataset_name)
            print("##################################")
            print("### "+dataset_name+" / "+anims_npz_filename)
            hpadataset = HPADataset( anims_npz_filename )
            loader = DataLoader( hpadataset, batch_size=4, shuffle=True, num_workers=0)
            print(str(hpadataset))
            # get some random training data
            dataiter = iter(loader)
            clips = dataiter.next()
            print("animation batch: type="+ str(type(clips)) +" size="+str(len(clips)))