"""
## Demo of Animation Dataset + Viewer
## 
"""
import math
import sys
import os
import numpy as np
from PIL import Image

import panda3d
from direct.gui.OnscreenText import OnscreenText
from direct.gui.OnscreenText import TextNode
from panda3d.core import NodePath
from direct.gui.OnscreenImage import OnscreenImage
from panda3d.core import PNMImage
from panda3d.core import Filename
from panda3d.core import Texture
from panda3d.core import CPTAUchar, PTAUchar, PTAStdfloat

import pydeepan
from pydeepan import log
from pydeepan.torch_import import *
from pydeepan import dataset as AnimDataset
from pydeepan.dataset.HPADataset import *
from pydeepan.viewer.PandaAnimViewer import PandaAnimViewer
from pydeepan.viewer.PandaSkeleton import HPAPandaSkeleton

from pydeepan.dataset.HPADataset import HPADataset






class HPADatasetDemoViewer(PandaAnimViewer):
    """
        HPADatasetDemoViewer: a viewer to illustrate the dataset usage in pytorch + display with panda/PandaAnimViewer

        Arguments:
            - txtDataset (pada stuff): text about the dataset
            - dataanim_txt_left (panda stuff): text displayed on the up left
            - styletxtnode (panda stuff): ???
            - dataset_names (array): list of dataset names, pressing 'd' change dataset to the next one
            - dataset_id (int): id of the dataset used, indice in the dataset_names array
            - dataset (HPADataset): the HPADataset (en héritage of Dataset from pytorch)
            - dataset_loader (loader): an instance of pytorch loader, specialised to HPADataset, with a list of transform define in the construction
            - styles_names (array of text): from import AnimDataset, the array styles_names[ self.dataset_name ]
            - anims (list): list of animations
            - labels (list): list of labels
            - labels_count (np.ndarray): number of animation per style, used to display animation in the end of each column
    """

    def __init__(self, nb_anim_batch=9, win_title="HPADatasetDemo", nb_anim_from_dataset=-1, dataset_names=None):
        PandaAnimViewer.__init__(self, AnyPandaSkeleton=HPAPandaSkeleton, animations=None, nb_anim_batch=nb_anim_batch, win_title=win_title)

        # text up left (info on dataset) and up right (info on style)
        self.txtDataset = OnscreenText(text="vide", pos=(-1.5, 0.9), scale=0.05, fg=(0.8, 0.2, 0.2, 1.0), mayChange=True, align=TextNode.ALeft)
        self.dataanim_txt_left = OnscreenText(text="vide", pos=(1.2, 0.9), scale=0.05, fg=(0.8, 0.2, 0.2, 1.0), mayChange=True, align=TextNode.ALeft)

        # text/style on the ground
        self.styletxtnode = render.attachNewNode('styletextnode_root')
        self.styletxt_onground_node = None  # will be created in nextDataset

        self.anims = []              # clips to be displayed
        self.labels = []             # real labels

        # dataset, see also nextDataset which do the load
        if dataset_names==None:
            self.dataset_names = [ "mixing", "styletransfer", "emilya", "MPI" ]
        else:
            self.dataset_names = dataset_names

        self.dataset_id = -1
        if nb_anim_from_dataset<0:
            nb_anim_from_dataset = nb_anim_batch
        self.nextDataset( nb_anim_from_dataset )

        # key
        self.addKey("d", "Animation: next dataset")


    @property
    def dataset_name(self):
        return self.dataset_names[ self.dataset_id ]


    @property
    def batch_from_dataset_size(self):
        return self.dataset_loader.batch_size


    def dataset_txt(self):
        txt = "dataset: " + self.dataset_name + "\n"
        txt += "batch size="+str(self.batch_size) + "\n"
        txt += "batch size displayed="+str(self.pandaskeletons_displayed_size) + "\n"
        txt += "batch size from dataset="+str(self.batch_from_dataset_size) + "\n"
        return txt


    def styles_txt(self):
        txt = "styles \n"
        for i, name in enumerate(self.styles_names):
            txt += str(i) + "=" + name + "\n"
        return txt


    def nextDataset(self, nb_anim_from_dataset=-1):
        """
            move to new dataset around: self.dataset_names (see constructor)
        """
        self.dataanim_id = np.array( [0,0,0] )
        self.dataset_id = (self.dataset_id+1) % len(self.dataset_names)

        # Load next Dataset
        lst_trans = transforms.Compose( [ HPATransformClipStyle() ] )             # a transform which will transform the data of the dataset into tuple<clip,style>
        self.dataset = HPADataset( AnimDataset.name2filename(self.dataset_name),  lst_trans)
        if nb_anim_from_dataset<0:
            nb_anim_from_dataset = self.batch_from_dataset_size
        self.dataset_loader = DataLoader( self.dataset, batch_size=nb_anim_from_dataset, shuffle=True, num_workers=0)
        print(str(self.dataset))

        # set IHM stuff
        self.txtDataset.text = self.dataset_txt()
        self.styles_names = AnimDataset.styles_names[ self.dataset_name ]
        assert len(self.styles_names) == self.dataset.dim_labels_style()
        self.dataanim_txt_left.text = self.styles_txt()
        print("style_names=" + str(self.styles_names))
        self.setDimOfDisplayBoard( len(self.styles_names), len(self.styles_names), 2.5, 2.5 )       # display board size is = number of styles, have to be done before nextBatch to put character in the right position
        self.nextBatch()

        if self.styletxt_onground_node is not None:
            for node in self.styletxt_onground_node:
                node.removeNode()
        self.styletxt_onground_node = np.empty( len(self.styles_names), dtype = NodePath)
        for i, name in enumerate(self.styles_names):
            text = TextNode("Styles_"+str(i))
            text.setText(name)
            self.styletxt_onground_node[i] = render.attachNewNode(text)
            self.styletxt_onground_node[i].setLightOff(1)
            self.styletxt_onground_node[i].setScale(0.9)
            self.styletxt_onground_node[i].setP( -90 )
            self.styletxt_onground_node[i].setR( -90 )
            pos = self.getPosXY(i,-2)
            self.styletxt_onground_node[i].setPos( pos[0], pos[1], pos[2]     ) #5, 10, 1)


    def nextBatch(self):
        """
           generate a new batch of animations.
        """
        # self.anims, self.labels, self.predicted, self.batch_reco_rate, self.id_list = self.dataset.get_batch()
        # get some random training data

        # Method 1: but you do not get the id of each samples
        # dataiter = iter(self.dataset_loader)
        # clips = dataiter.next()
        # self.anims, self.labels = clips[0], clips[1]
        # self.anims =  self.dataset.postprocess( self.anims.cpu().numpy() )
        # print("animation batch: type="+ str(type(clips)) +" size="+str(len(clips)))

        # Method 2
        i_list = next(iter(self.dataset_loader.batch_sampler))
        self.anims = []              # clips to be displayed
        self.labels = []             # real labels
        for i in i_list:
            clip = self.dataset.get_brut_clip(i)         # clip ready to be displayed
            _, label = self.dataset[i]
            self.labels.append( label )
            self.anims.append( clip )
        self.anims = np.stack( self.anims )
        self.labels = np.stack( self.labels )

        # Poisition anim + color + style label
        self.labels_count = np.zeros( [self.dataset.dim_labels_style(),], dtype=int)
        for i in range(self.batch_from_dataset_size):
            ske = self.pandaskeletons[i]
            ske.setColorStyle(3)
            self.setAnimationPosXY( i, self.labels[i],  self.labels_count[ self.labels[i] ])
            self.labels_count[ self.labels[i] ] += 1
            txtadd = self.styles_names[ self.labels[i] ]
            txtadd = ""
            ske.setId( i_list[i], txtadd )
        self.txtDataset.text = self.dataset_txt()
        self.setPosesOfAllSkeletons()


    def keyHandler(self,key):
        if key=="d":
            self.nextDataset()
        elif key=="=":
            super().keyHandler(key)
            self.txtDataset.text = self.dataset_txt()
        elif key=="-":
            super().keyHandler(key)
            self.txtDataset.text = self.dataset_txt()
        else:
            super().keyHandler(key)


    def animate(self):
        super().animate()




if __name__ == '__main__' and sys.flags.interactive == 0:

    log_file = pydeepan.dir_pydeepan + "data/HPADatasetDemo.log"
    with log.create_logger(log_file, True):
        app = HPADatasetDemoViewer( nb_anim_batch=9 )
        app.run()







# if __name__ == "__main__":

#     anims_npz_filename = dataset.name2filename('styletransfer')

#     log_file = pydeepan.dir_pydeepan + "data/hpadataset.log"
#     log = log.create_logger(log_file)

#     hpadataset = HPADataset( anims_npz_filename )
#     loader = DataLoader( hpadataset, batch_size=4, shuffle=True, num_workers=0)

#     # get some random training data
#     dataiter = iter(loader)
#     clips = dataiter.next()
#     print("animation batch: type="+ str(type(clips)) +" size="+str(len(clips)))
#     if True:
#         clipsnp = hpadataset.postprocess( clips['clip'].numpy() )
#         print("animation batch numpy: type="+ str(type(clipsnp)) +" shape="+str(clipsnp.shape))
#         #HPAPlotViewer.HPAPlotDraw( clipsnp )
#         PandaAnimViewer.HPAPandaDraw( clipsnp )
