'''Train an autoencoder on animation (Holden format called HPA=Holden Positions Animation, see readme.md)

    TRAINING
    cd .../deepan/pydeepan
    python encoder/AutoEncoder.py --anims cmu --load none --save ../data/dl_models/autoencoder_animation_cmu.pth --loop 1000000 -lr 0.0001 --batch 8 --device gpu

    TEST autoencoder :
    python encoder/AutoEncoder.py --anim styletransfer --batch 2 --loop 0 --load ../data/dl_models/autoencoder_animation_cmu.pth --show
'''

import torch

import pydeepan
from pydeepan import log, tools
from pydeepan.torch_import import *
from pydeepan import dataset
from pydeepan.latent.IdentityEncoder import IdentityEncoder





class AutoEncoderNet(nn.Module, IdentityEncoder):
    """
    ############### Auto Encoder
    #   Network(
    #    DropoutLayer(amount=dropout, rng=rng),
    #    Conv1DLayer(filter_shape=(256, 73, 25), input_shape=(batchsize, 73, window), rng=rng),
    #    BiasLayer(shape=(256, 1)),
    #    ActivationLayer(),         # default = RELU
    #    Pool1DLayer(input_shape=(batchsize, 256, window)),  # default arg=> max pool of 2
    #),
    #Network(
    #    Depool1DLayer(output_shape=(batchsize, 256, window), depooler='random', rng=rng),
    #    DropoutLayer(amount=dropout, rng=rng),
    #    Conv1DLayer(filter_shape=(73, 256, 25), input_shape=(batchsize, 256, window), rng=rng),
    #    BiasLayer(shape=(73, 1))
    """

    def __init__(self, device=torch.device("cpu"), justEncode=False, dropout=0.25, depooler='random'):
        nn.Module.__init__(self)
        # input: batchsize x (channel=73, window(=240=time dimension))  ~ signal 1D de taille 240 frames avec 73 channels
        # p=probability of dropout
        self.dropout = nn.Dropout(p=dropout)
        # (73 channels, 256 conv to apply, filterSize=25) => batchsize x (73 channels x 256 conv x 240-25+1 for time)
        self.conv = nn.Conv1d(73, 256, 25, bias=True)
        # => output: (batchsize, 73 Channels, 256=nb conv, timeDim/2=120)
        self.maxpool = nn.MaxPool1d(2, return_indices=True)
        # => output: (batchsize, 73 Channels, 256=nb conv, 240 time)

        self.unpool = nn.MaxUnpool1d(2)
        self.deconv = nn.ConvTranspose1d(256, 73, 25, bias=True)
        self.justEncode = justEncode
        IdentityEncoder.__init__(self, [73, 240], device)
        self.to( self.device )


    def forward(self, x):
        """
            Process a batch of anims x.
            x has to be a 'batch of anims' in pytorch-tensor preprocessed (CPU or CUDA)
        """
        self.to( self.device )                  # envoie le net autoencoder sur le device
        code, ind = self.encode(x, True)
        if self.justEncode:
            return code
        output = self.decode(code, ind)
        return output


    def encode(self, x:torch.tensor, return_ind=False):
        """
            Encode a batch of anims x.
            'batch of anims' has to be pytorch-tensor preprocessed.
        """
        #print("inputs="+str(x.size()))
        #### ENCODE x = self.encoder(x)
        x = self.dropout(x)
        x = self.conv(x)
        x = F.relu(x)
        x, ind = self.maxpool(x)
        #print("encoded="+str(x.size()))
        if return_ind:
            return x, ind
        else:
            return x


    def decode(self, x:torch.tensor, ind:torch.tensor):
        """
            Decode a batch of anims x.
            'batch of anims' has to be pytorch-tensor preprocessed.
        """
        #### DECODE x = self.decoder(x)
        x = self.unpool(x, indices=ind)
        x = F.relu(x)
        x = self.dropout(x)
        x = self.deconv(x)
        #print("decoded="+str(x.size()))
        return x


    def __call__(self,x):
        """
            Process a batch of anims. The batch of anims could be x, or a list with x like that x=[batch of anims, style] or tuple x=(batch of anims, style)
            But 'batch of anims' has to be pytorch-tensor preprocessed (see HPADataset).
        """
        if isinstance(x,list):
            x0 = x[0]
            x0 = x0.to(self.device)
            return [ self.forward(x0), x[1] ]
        elif isinstance(x,tuple):
            x0 = x[0]
            sq = False
            if x0.dim()==2:
                sq = True
                x0 = x0.unsqueeze(0)
            x0 = x0.to(self.device)
            y0 = self.forward(x0)
            if sq:
                y0 = y0.squeeze()
            return ( y0, x[1] )
        else:
            #x.to(self.device)
            return self.forward(x)


    def __str__(self):
        """
            return a text describing the autoencoder net
        """
        res = "=============================== AutoencoderNet =================================================================\n"
        res += super(AutoEncoderNet,self).__str__() + "\n"
        s = StringIO()
        stdout_save = sys.stdout
        sys.stdout = s
        summary(self, tuple(self.input_shape), device=self.device.type)
        sys.stdout = stdout_save  #sys.__stdout__
        res += s.getvalue()
        res += "device="+str(self.device)+"\n"
        res += "training=" + str(self.training)+"\n"
        res+= "input_shape=" + str(self.input_shape) + "\n"
        res += "===============================================================================================================\n"
        return res






class AutoEncoder:
    """
        AutoEncoder = an AutoEncoderNet + a Dataset
    """

    def __init__(self, dataset_npz_filename, batch_size, device_name, dropout=0.25, depooler='random'):
        self.hpa_transform = transforms.Compose([
                    dataset.HPADataset.HPATransformClipWithoutLabel()
                ])
        self.hpadataset = dataset.HPADataset.HPADataset(dataset_npz_filename, self.hpa_transform)
        self.loader = DataLoader(self.hpadataset, batch_size=batch_size, shuffle=True, num_workers=0)
        device = torch.device(device_name)
        self.net = AutoEncoderNet( device, False, dropout, depooler)
        self.dataiter = iter(self.loader)
        clips = self.dataiter.next()
        print("data shape from loader="+str(clips.shape))


    @staticmethod
    def filenames(datasetname, best=False):
        ae_name = "autoencoder_animation"
        npz_anims_file = dataset.name2filename(datasetname)
        if npz_anims_file != datasetname and npz_anims_file != "":
            ae_name += "_"+datasetname
        if best:
            ae_name += "_best"
        ae_file = pydeepan.dir_pydeepan_dlmodels + ae_name + ".pth"
        return ae_file


    @staticmethod
    def expand_args(args):
        args.anims = dataset.name2filename(args.anims)
        if args.load == "default":
            args.load = AutoEncoder.filenames("cmu", True)
        if not hasattr(args,'save') or args.save == "default":
            args.save = AutoEncoder.filenames(args.anims, False)
        args.device = ("cuda" if torch.cuda.is_available() and (
            args.device == "default" or args.device == "gpu") else "cpu")
        if args.log == "default":
            args.log = args.save + ".log"
        if args.show == "no":
            args.show = False
        else:
            args.show = True


    def __call__(self, anim, need_preprocess=True):
        # if need_preprocess: take a numpy, preprocess it; else take a torch_tensor. Return a numpy
        if need_preprocess:
            a = self.hpadataset.preprocess(anim)
            a = self.hpadataset.to_torch_tensor(a)
        else:
            a = anim
        a.to( self.net.device)
        a = self.net(a)
        a = a.detach().cpu().numpy()
        return self.hpadataset.postprocess(a)


    def __str__(self):
        res = "AutoEncoder:\n"
        res += "===> net:\n" + self.net.__str__()
        res += "===> net is_cuda=" + str(next(self.net.parameters()).is_cuda) + "\n"
        res += "===>HPAdataset:\n" + self.hpadataset.__str__()
        return res


    def load(self, filename):
        #self.net = AutoEncoderNet
        #self.net = torch.load(filename, )
        self.net.load_state_dict(  torch.load(filename) )
        #self.net = self.net.to(self.net.device)
        #self.hpadataset.load_mean_std( filename )


    def test_dataset(self, show):
        # get some random training data, just to see if it works
        print(self.hpadataset)
        #dataiter = iter(self.loader)
        clips = self.dataiter.next()
        print("animation batch: type=" + str(type(clips)) + " size="+str(len(clips)))
        if show:
            clipsnp = self.hpadataset.postprocess(clips.numpy())
            print("animation batch numpy: type=" +
                  str(type(clipsnp)) + " shape="+str(clipsnp.shape))
            #PlotDraw(clipsnp)


    def test_and_print(self):
        clips = self.dataiter.next()
        clips = clips.to(self.net.device)
        print("animation batch: type=" + str(type(clips)) + " size="+str(len(clips)))
        clips_np_encdec = self.__call__(clips, need_preprocess=False)
        clips_np = clips.detach().cpu().numpy()  # .swapaxes( clips.detach().numpy(), 1, 2)
        clips_np = self.hpadataset.postprocess(clips_np) #np.swapaxes( clips_np, 1, 2)
        command_line = np.concatenate(  (clips_np, clips_np_encdec)  )
        print("animation batch numpy: type=" + str(type(command_line)) + " shape="+str(command_line.shape))
        #PlotDraw(command_line)
        #HPAPlotViewer.HPAPlotDraw( command_line )


    def train(self, loop=10, learningrate=0.0001, filename="ae.pth"):
        ############ TRAINNING
        #self.hpadataset.save_mean_std( filename )
        if loop == 0:
            return
        ## SGD config: Stochastic Gradient Descent Config
        criterion = nn.MSELoss()
        optimizer = optim.SGD(self.net.parameters(),
                              lr=learningrate, momentum=0.9)
        print("loader size="+str(len(self.loader)))
        self.net = self.net.to(self.net.device)
        ## train
        for epoch in range(loop):  # loop over the dataset multiple times
            running_loss = 0.0
            nitem_loss = 0
            for i, data in enumerate(self.loader, 0):         # mini batch
                inputs = data
                desiredouputs = inputs
                inputs, desiredouputs = inputs.to(self.net.device), desiredouputs.to(self.net.device)

                # zero the parameter gradients
                optimizer.zero_grad()

                # forward + backward + optimize
                outputs = self.net(inputs)
                loss = criterion(outputs, desiredouputs)
                loss.backward()
                optimizer.step()

                # statistics
                running_loss += loss.item()
                nitem_loss += 1

            ############ SAVE
            print('[%d, %5d] loss: %.3f' %
                  (epoch + 1, i + 1, running_loss / nitem_loss))
            print("save..."+str(filename))
            torch.save(self.net.state_dict(), filename)
            # dummy_input = inputs
            # torch.onnx.export(self.net, dummy_input, filename + ".onnx")

        filename = tools.add_to_filename(
            filename,  "_loss"+str(float(round(running_loss / nitem_loss, 3))))
        print("save..."+str(filename))
        torch.save(self.net.state_dict(), filename)
        print('Finished Training')




if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-sh", "--show", help="show/vizualize some examples of animations (yes or no)", type=str, default="yes")
    parser.add_argument(
        "-l", "--load", help="load a model: 'none' or 'default' or a file name", type=str, default="none") #"default")
    parser.add_argument(
        "-s", "--save", help="filename of the model for saving: 'default' or a file name", type=str, default="default")
    parser.add_argument(
        "-g", "--log", help="filename for the log: 'default' or a file name", type=str, default="default")
    parser.add_argument(
        "-a", "--anims", help="animation database for loading (specify 'cmu', 'styletransfer' or path to the npz file)", type=str, default="cmu") #"styletransfer")
    parser.add_argument(
        "-lo", "--loop", help="loop over dataset (0 for no training)", type=int, default=1000000)
    parser.add_argument(
        "-d", "--device", help="cpu or gpu or default", type=str, default="default")
    parser.add_argument("-b", "--batch", help="batch size", type=int, default=16)
    parser.add_argument("-lr", "--learningrate", help="learning rate", type=float, default=0.001)
    args = parser.parse_args()


    ### Just to test and show a pretrained autoencoder
    # python encoder/AutoEncoder.py --anim styletransfer --loop 0 --load ../data/dl_models/autoencoder_animation_cmu.pth --show
    if False: #True: #False:
        print("====> WARNING WARNING: ARG are tweaked just to run/debug test_and_show ====================================")
        print("====> WARNING WARNING: ARG are tweaked just to run/debug test_and_show ====================================")
        print("====> WARNING WARNING: ARG are tweaked just to run/debug test_and_show ====================================")
        print("====> WARNING WARNING: ARG are tweaked just to run/debug test_and_show ====================================")
        print("====> WARNING WARNING: ARG are tweaked just to run/debug test_and_show ====================================")
        args.show = "yes"
        args.load = "data/dl_models/autoencoder_animation_cmu_best.pth"
        args.save = "data/dl_models/autoencoder_animation_cmu_test.pth"
        args.anim = "styletransfer"
        args.device = "gpu"
        args.batch = 2
        args.loop = 10


    AutoEncoder.expand_args(args)

    with log.create_logger(args.log, True):
        command_line = ""
        for i in sys.argv:
            command_line += i+" "
        print("=================================================================== ")
        print("====> " + command_line)
        print("=================================================================== ")
        print("====> data animations file='"+args.anims+"'")
        print("====> loop over dataset (0 means no training)="+str(args.loop))
        print("====> learning rate="+str(args.learningrate))
        print("====> batch_size="+str(args.batch))
        print("====> show some example of animations (yes/no): "+str(args.show))
        print("====> load the model '"+args.load+"'")
        print("====> save the model '"+args.save+"'")
        print("====> log file '"+args.log+"'")
        print("====> name of device="+args.device)
        print("=================================================================== ")


        ############# NETWORK definition/configuration
        net = AutoEncoder(args.anims, args.batch, args.device)
        print("===net===")
        print(net)

        if args.load != "none":
            print("####################################")
            print("############### loading..."+args.load)
            print("####################################")
            print("")
            net.load(args.load)
            #net = net.to(net.device)
            print(str(net))

        if args.loop > 0:
            print("####################################")
            print("###############  train...")
            print("####################################")
            net.train(args.loop, args.learningrate, args.save)

        if args.show:
            print("####################################")
            print("###############  test and show...")
            print("####################################")
            net.test_and_print()

        print("END")
