
import math
import sys
import os
import numpy as np

import panda3d
from direct.gui.DirectGui import *

from pydeepan import dataset
from pydeepan.viewer.PandaViewer import PandaViewer
from pydeepan.viewer.PandaSkeleton import HPAPandaSkeleton





class PandaAnimViewer(PandaViewer):
    """
        Base Class to Display Batch of Animations
        Arguments:
            - anims(np.ndarray): store all animations, type of animation are generic but the array of animations must be an ndarray
            - pandaskeletons (np.ndarray): all skeleton to be displayed (see skeleton_type to have the class type of them)
            - skeleton_type (class type): PandaSkeleton is an example of skeleton_type. You can pass any other skeleton class (like a template parameter).
            - batch_size(int): number of animations in batches, not all animations are displayed, only a batch is displayed
            - pandaskeleton_displayed_size(int): number of displayed animations, must be < batch_size. Even in the batch, user can reduce the number of visible animations
            - dimx, dimy (int): size of the square to display animations on the ground
            - paddingx, paddingy (float): space/padding between animations on the ground
            - anim_id (int): id in self.anims of the first animations of the batch. Here the batch is a sequence from anim_id to anim_id+batch_size-1
            - frame_id: frame to be displayed, when animation is running, this value is incremented each frame
            - displayed_anim_id (bool): indicate if text/anim_id is displayed on the ground near the skeleton
    """
    def __init__(self, AnyPandaSkeleton, animations=None, nb_anim_batch=25, win_title="AnimPandaViewer", *therest):
        """
            nb_anim_batch: indicate the number of animations in a batch
        """
        PandaViewer.__init__(self, win_title=win_title, *therest)
        self.skeleton_type = AnyPandaSkeleton
        self.axis3D.detachNode()
        self.anim_id = 0
        self.frame_id = 0
        self.display_anim_id = True

        self.anims = animations
        length = int(math.sqrt( nb_anim_batch ))
        self.setDimOfDisplayBoard(length,length,2.5, 2.5)         # square of skeleton length*2.5  X  length*2.5

        self.pandaskeletons = None
        self.pandaskeletons_displayed_size = 0
        self.setCharacterBatchSize( nb_anim_batch )          # it create the self.pandaskeleton and init self.pandaskeleton_displayed_size

        self.addKey("o", "Animation: previous batch of animations")
        self.addKey("p", "Animation: Next batch of animations")
        self.addKey("i", "Animation: toggle animation id on the ground")
        self.addKey("&", "Animation: decrease batch size to 1")
        self.addKey("-", "Animation: decrease batch size")
        self.addKey("=", "Animation: increase batch size")

        # Add button
        #self.button = DirectButton(text=("Quit"), pos=(-1.4, 0, 0.95), scale=.05, command=self.setText)


    # def setText(self):
    #     print("button clicked")
    #     self.keyHandler("q")


    def setChacacterDisplayedSize(self, nb_anim_showed):
        """
         Set the number of animation really displayed. Not all batch are necessary displayed
         The function iterate on hapdraw (size=batch_size) and attach or not to the graph
         This function has to be called after setCharacterBatchSize which create all the pandaskeleton
        """
        if self.pandaskeletons is not None:
            if nb_anim_showed>self.batch_size:
                nb_anim_showed = self.batch_size
            if nb_anim_showed<1:
                nb_anim_showed = 1
            self.pandaskeletons_displayed_size = nb_anim_showed
            for i in range(self.batch_size):
                if i<nb_anim_showed:
                    self.pandaskeletons[i].skeleton.reparentTo(render)
                else:
                    self.pandaskeletons[i].skeleton.detachNode()


    def setCharacterBatchSize(self, nb_anim_batch):
        """
            Create all the pandaskeletons (remove them if already exist)
        """
        if self.pandaskeletons is not None:
            for i in self.pandaskeletons:
                i.skeleton.removeNode()
        self.pandaskeletons = np.empty( nb_anim_batch, dtype = self.skeleton_type)
        for i in range( self.batch_size):
            self.pandaskeletons[i] = self.skeleton_type(self)
            self.pandaskeletons[i].setId( i )
            self.setAnimationPosXYFromId(i)
        self.pandaskeletons_displayed_size = self.batch_size
        if self.anims is not None:
            self.setPosesOfAllSkeletons()


    @property
    def batch_size(self):
        """ return the number of animations in a batch """
        return self.pandaskeletons.shape[0]


    def setDimOfDisplayBoard(self, dimx, dimy, paddingx=2.5, paddingy=2.5):
        self.dimx = dimx
        self.dimy = dimy
        self.paddingx = paddingx
        self.paddingy = paddingy


    @property
    def minx(self):
        return -self.paddingx*self.dimx*0.5


    @property
    def miny(self):
        return -self.paddingy*self.dimy*0.5


    def getPosXY(self, x, y):
        return [self.minx+(x+0.5)*self.paddingx, self.miny+(y+0.5)*self.paddingy, 0]


    def setAnimationPosXY(self, anim_id, x, y):
        """ Positionne l'animation anim_id en (x,y) """
        self.pandaskeletons[anim_id].translate = self.getPosXY(x,y)


    def setAnimationPosXYFromId(self, id):
        x = id % self.dimx
        y = int(id / self.dimx)
        self.setAnimationPosXY(id, x, y)


    def nextBatch(self):
        """ extract the next batch from the dataset """
        self.anim_id = (self.anim_id + self.batch_size) % self.anims.shape[0] 
        for i in range( self.batch_size ):
            id = (self.anim_id+i)%self.anims.shape[0]
            self.pandaskeletons[i].setId( id )
        self.setPosesOfAllSkeletons()
        self.printAnimInfo()


    def previousBatch(self):
        """ extract the previous batch from the dataset """
        self.anim_id = (self.anim_id - self.batch_size) % self.anims.shape[0] 
        self.setPosesOfAllSkeletons()
        self.printAnimInfo()


    def setPosesOfAllSkeletons(self): #, frame_id=-1):
        # if frame_id>=0:
        #     self.frame_id = frame_id
        for i in range( self.batch_size ):
            id = (self.anim_id+i)%self.anims.shape[0]
            self.pandaskeletons[i].setPose(  self.anims[ id  ][ self.frame_id] )  # animation 0, pose 0 => 73 dim
            #self.pandaskeletons[i].setPose(  self.anims[ id ], self.frame_id )  # animation 0, pose 0 => 73 dim
            self.pandaskeletons[i].setDisplayAnimId( self.display_anim_id )


    def keyHandler(self,key):
        if key=="b":
            if self.frame_id>0:
                self.frame_id -= 1
            else:
                self.frame_id = self.anims.shape[1]-1
            self.printAnimInfo()
            self.setPosesOfAllSkeletons()
        elif key=="o":
            self.previousBatch()
        elif key=="p":
            self.nextBatch()
        elif key=="i":
            self.printAnimInfo()
            self.display_anim_id = not self.display_anim_id
            self.setPosesOfAllSkeletons()
        elif key=="&":
            self.setChacacterDisplayedSize(1)
            print("displayed_size="+str(self.pandaskeletons_displayed_size) + " batch=size="+str(self.batch_size))
        elif key=="=":
            self.setChacacterDisplayedSize(self.pandaskeletons_displayed_size+1)
            print("displayed_size="+str(self.pandaskeletons_displayed_size) + " batch=size="+str(self.batch_size))
        elif key=="-":
            self.setChacacterDisplayedSize( self.pandaskeletons_displayed_size-1 )
            print("displayed_size="+str(self.pandaskeletons_displayed_size) + " batch=size="+str(self.batch_size))
        else:
            super().keyHandler(key)


    def printAnimInfo(self):
            print("animation info: anim_id="+str(self.anim_id) + "/" + str(self.anims.shape[0])+ "  frame_id="+str(self.frame_id)+"/"+str(self.anims.shape[1]))


    def animate(self):
        self.frame_id = (self.frame_id + 1) % self.anims.shape[1]
        self.setPosesOfAllSkeletons()




def HPAPandaDraw(anims):
    app = PandaAnimViewer( HPAPandaSkeleton, anims)
    app.run()



if __name__ == '__main__' and sys.flags.interactive == 0:

    #database = np.load( dataset.dir_anims_processed_npz + 'holden/data_styletransfer.npz')
    #database = np.load( dataset.name2filename("styletransfer") )
    #database = np.load( dataset.name2filename("emilya") )
    database = np.load( dataset.name2filename("cmu") )
    anims = database['clips']
    app = PandaAnimViewer( HPAPandaSkeleton, anims)
    app.run()
