import numpy as np
import panda3d
from direct.gui.OnscreenText import TextNode

from pydeepan.chara.Skeleton import Skeleton




class PandaSkeleton:
    """ 
        class to draw a skeleton. Unit should be around an human in meter (between 1.50 and 2 units) 
        Arguments:
            - viewer (Panda-Based Viewer): the viewer that displayes the PandaSkeleton
            - n_joints (int): number of joints
            - translate (panda transform): translate the skeleton in the world
            - scale (panda scale): scale the skeleton
            - skeleton (NodePath): the Panda Node, root of the displayed skeleton.
            - joints (np.ndarray of NodePath): all joints of the skeleton
            - cylinders (np.ndarray of NodePath): all cylinder of the skeleton
            - parent_id (np.ndarray of int): parent id of each joints. Someone has to fill in this array. 
            - sphere_radius
            - cylinder_radius
            - id (int): the id of the skeleton which could be displayed on the ground (for instance the id in the dataset)
            - idtxt (text): version text of the id
    """
    def __init__(self, viewer, n_joints, translate= [0,0,0], scale=1.0, id=-1):
        self.viewer = viewer
        self.n_joints = n_joints
        self.translate = translate
        self.scale = scale
        self.joints = np.empty( [self.n_joints, ], dtype=panda3d.core.NodePath)         # sphere of each joint
        self.cylinders = np.empty( [self.n_joints, ], dtype=panda3d.core.NodePath)      # cylinder of each limb

        self.parent_id = np.zeros( [self.n_joints, ], dtype=int)        # indice of the parent, -1 means no parent. It is used to place the cylinder
        #self.parent_id[0] = -1             # this remove the creation of root's cylinder

        self.skeleton = render.attachNewNode("skeleton")
        self.setSkeletonTransform()

        # create an empty skeleton: sphere + cylinders toward parent (if parent exits)
        self.sphere_radius = scale * 6.0
        self.cylinder_radius = scale * 4.0
        for i in range(self.n_joints):      # each node has a cylinder, even the root node (it will be detached later)
            self.joints[i] = self.skeleton.attachNewNode("joint_sphere")
            self.joints[i].setPos( 0, 0, i)
            self.joints[i].setScale( self.sphere_radius, self.sphere_radius, self.sphere_radius )
            viewer.sphere.instanceTo(self.joints[i])
            self.cylinders[i] = self.skeleton.attachNewNode("joint_cylinder_toward_parent")
            if self.parent_id[i]>=0:
                viewer.cylinder.instanceTo(self.cylinders[i])

        self.createIdTxtNode()
        self.setId( id )

        self.setColorStyle()


    def createIdTxtNode(self):
        self.id_txt = TextNode("id")
        self.id_txt.setText("id")
        self.id_txt_node = self.joints[0].attachNewNode(self.id_txt)
        self.id_txt_node.setPos( 1, 0, 0)
        self.id_txt_node.setLightOff(1)
        self.id_txt_node.setScale(10.0)
        self.id_txt_node.setP( -180 )


    def setId(self, id, txt=""):
        self.id = id
        t = str(self.id)
        if txt!="":
            t = t+"/"+txt
        self.id_txt.setText(t)


    def setDisplayAnimId(self, b):
        if not b:
            self.id_txt_node.detachNode()
        else:
            self.id_txt_node.reparentTo( self.joints[0])


    def setSkeletonTransform(self):
        self.skeleton.setPos( self.translate[0], self.translate[1], self.translate[2] )
        self.skeleton.setP(90)
        self.skeleton.setScale( self.scale, self.scale, self.scale)


    def isEnable(self):
        return self.skeleton.has_parent()


    def enable(self):
        self.skeleton.reparentTo(render)


    def disable(self):
        self.skeleton.detachNode()


    def setColor(self, rsph,gsph,bsph,asph, rcyl, gcyl, bcyl, acyl ):
        for i in range(self.n_joints):
            self.joints[i].setColor( rsph,gsph,bsph,asph)
            if self.parent_id[i]>=0:
                self.cylinders[i].setColor( rcyl, gcyl, bcyl,acyl)


    def setColorStyle(self, style=-1):
        """ 
            -1: neutral, near blue
            0: wrong, near red
            1: good, near green
            2: transparency 
            3: light
        """
        if style==0:
            self.setColor( 0.8, 0.0, 0.3, 1.0,  0.7, 0.1, 0.6, 1.0)             # wrong
        elif style==1:
            self.setColor( 0.2, 0.7, 0.4, 1.0,  0.3, 0.8, 0.7, 1.0)             # good
        elif style==2:                      
            self.setColor( 0.0, 0.7, 0.9, 1.0,  0.0, 0.7, 0.9, 1.0)             # ???
        elif style==3:                      
            self.setColor( 0.7, 0.7, 1.0, 1.0,  0.7, 0.75, 1.0, 1.0)             # light
        else:                               
            self.setColor( 0.5, 0.5, 0.8, 1.0,  0.7, 0.7, 0.9, 1.0)             # neutral


    def setPose(self, ske : Skeleton):
        assert ske.n_joints == self.n_joints
        for i in range(ske.n_joints):
            joint = ske[i].transform.translation
            if not ske[i].is_root:
                Pi = ske[i].transform.translation                # Position joint i
                Pp = ske[i].parent.transform.translation         # Position parent of joint i
                Pi = Point3( Pi.x, Pi.y, Pi.z)                   # glm.vec3 to Panda.Point3
                Pp = Point3( Pp.x, Pp.y, Pp.z)                   # glm.vec3 to Panda.Point3
                self.viewer.placeCylinder( self.cylinders[i], Pi, Pp, self.cylinder_radius )
            else:
                self.cylinders[i].detachNode()







class HPAPandaSkeleton(PandaSkeleton):
    """
        # ==========================================================================================================
        # ** Holden's format (in the NPZ files) : 73 values = 22 Positions(XYZ) + VX, VZ, VangY + 4 contacts
        #  0 ** 0, 1, 2 : projection du root sur le sol ?       -1 (father) 
        #  1 ** 3, 4, 5 : root; vertebre basse                  -1
        #  2 ** 6, 7, 8 : hanche gauche                         1
        #  3 ** 9, 10, 11 : genou gauche                        2
        #  4 ** 12, 13, 14 : cheville gauche                    3
        #  5 ** 15, 16, 17 : orteils gauche                     4
        #  6 ** 18, 19, 20 : hanche droite                      1
        #  7 ** 21, 22, 23 : genou droite                       6
        #  8 ** 24, 25, 26 : cheville droite                    7
        #  9 ** 27, 28, 29 : orteils droite                     8
        # 10 ** 30, 31, 32 : vertebre milieu                    1
        # 11 ** 33, 34, 35 : vertebre haute                     10
        # 12 ** 36, 37, 38 : nuque                              11
        # 13 ** 39, 40, 41 : tete                               12
        # 14 ** 42, 43, 44 : epaule gauche                      12
        # 15 ** 45, 46, 47 : coude gauche                       14
        # 16 ** 48, 49, 50 : poignet gauche                     15
        # 17 ** 51, 52, 53 : doigts gauche                      16
        # 18 ** 54, 55, 56 : epaule droite                      12
        # 19 ** 57, 58, 59 : coude droite                       18
        # 20 ** 60, 61, 62 : poignet droite                     19
        # 21 ** 63, 64, 65 : doigts droite                      20
        # xx ** 66 : Vitesse X
        # xx ** 67 : Vitesse Z
        # xx ** 68 : Vitesse angulaire autour de Y
        # xx ** 69, 70 : Contact pied gauche
        # xx ** 71, 72 : Contact pied droit
        # ==========================================================================================================

    """

    def __init__(self, viewer, translate=[0,0,0]):
        PandaSkeleton.__init__(self, viewer=viewer, n_joints=22, translate=translate, scale=0.1)         # 22 nodes in the HPA format

        #                           0   1  2  3  4  5  6  7  8  9 10  11  12  13  14  15  16  17  18  19  20  21
        self.parent_id = np.array([-1, -1, 1, 2, 3, 4, 1, 6, 7, 8, 1, 10, 11, 12, 12, 14, 15, 16, 12, 18, 19, 20])
        #      parents = np.array([-1, 0, 1, 2, 3, 0, 5, 6, 7, 0, 9, 10, 11, 11, 13, 14, 15, 11, 17, 18, 19])

        #self.cylinders[0].detachNode()

    @property
    def pose_dim(self):
        return 73


    def setPose(self, pose_np:np.ndarray):
        """
        """
        assert type(pose_np) == np.ndarray
        self.setSkeletonTransform()
        for i in range( self.n_joints): 
            j = self.joints[i]
            j.setPos( pose_np[ 3*i ], pose_np[3*i+1], pose_np[3*i+2] )
            if self.parent_id[i]>=0:
                pid = self.parent_id[i]             # id parent
                Pi = self.joints[i].getPos()        # Position joint i
                Pp = self.joints[pid].getPos()      # Position parent of joint i
                self.viewer.placeCylinder( self.cylinders[i], Pi, Pp, self.cylinder_radius )
            else:
                self.cylinders[i].detachNode()




